#!/usr/bin/env python
'''
dicom sort rule functions:

    sort_rule_demo: a simple demo sort rule
    sort_rule_CFMM: CFMM's sort rule

Author: YingLi Lu
Email:  yinglilu@gmail.com
Date:   2018-05-22

note:
    Tested on windows 10/ubuntu 16.04, python 2.7.14
'''

import os
import re
import pydicom
import logging
import csv
import dcmstack as ds


def sort_rule_demo(filename):
    '''
    A simple sort rule:

    patient_name
      |-study_date
        |-series_number
          |-{patient_name}.{study_data}.{series_number}.{image_instance_number:04d}.dcm
          ... 
        |-series_number
        ...

    intput:
        filename: dicom filename
    output:
        a dictionary:
            key: filename
            value: patient_name/study_date/sereis_number/{patient_name}.{study_data}.{series_number}.{image:04d}.dcm

    '''
    logger = logging.getLogger(__name__)

    def clean_path(path):
        return re.sub(r'[^a-zA-Z0-9.-]', '_', '{0}'.format(path))

    try:
        dataset = pydicom.read_file(filename, stop_before_pixels=True)

        patient_name = clean_path(dataset.PatientName.replace('^', '_'))
        #print('patient_name', patient_name)
        study_date = clean_path(dataset.StudyDate)
        series_number = clean_path(
            '{series_number:04d}'.format(series_number=dataset.SeriesNumber))

        path = os.path.join(patient_name, study_date, series_number)
        sorted_filename = '{patient}.{study_date}.{series_number}.{image_instance_number:04d}.dcm'.format(
            patient=patient_name,
            study_date=study_date,
            series_number=dataset.SeriesNumber,
            image_instance_number=dataset.InstanceNumber,
        )
        sorted_filename = clean_path(sorted_filename)

    except Exception as e:
        logger.exception('something wrong with {}'.format(filename))
        logger.exception(e)
        return None

    sorted_full_filename = os.path.join(path, sorted_filename)
    return sorted_full_filename


def sort_rule_CFMM(filename, args):
    '''
    CFMM's Dicom sort rule

    intput:
        filename: dicom filename
    output:
        a dictionary:
            key: filename
            value: pi/project/study_date/patient/studyID_and_hash_studyInstanceUID/series_number
                   /{patient}.{modality}.{study}.{series:04d}.{image:04d}.{date}.{unique}.dcm

    CFMM's DICOM data Hierarchical structure: (same with CFMM's dcmrcvr.https://gitlab.com/cfmm/dcmrcvr)
    root_dir/
        -PI->first part of StudyDescription: John^Project.
            -project ->second part of StudyDescription: John^3T_Project.
                -19700101 ->StudyDate
                    -1970_01_01_C001 ->patientName
                    -1.AC168B21 -> dataset.StudyID + '.' + hashcode(dataset.StudyInstanceUID)
                            -0001->series number
                            -0002
                            -0003
                            -0004
                            -0005
                            -0006
                            -0007
                            ...
                    -1970_01_01_C002
                        -1.AC168B24
                            ...
                    -1970_01_01_C003
                        -1.AC168B3C
    '''

    logger = logging.getLogger(__name__)

    def clean_path(path):
        return re.sub(r'[^a-zA-Z0-9.-]', '_', '{0}'.format(path))

    def hashcode(value):
        code = 0
        for character in value:
            code = (code * 31 + ord(character)) & 0xffffffff
        return '{0:08X}'.format(code)

    try:
        dataset = pydicom.read_file(filename, stop_before_pixels=True)

        # StudyDescription
        # CFMM's newer data:'PI^project'->['PI','project']
        # CFMM's older GE data:'PI project'->['PI','project']
        if 'StudyDescription' in dataset and len(dataset.StudyDescription) > 0:
            StudyDescription = dataset.StudyDescription
        else:
            StudyDescription = args.StudyDescription

        pi_project = StudyDescription.replace('^', ' ').split()
        pi = clean_path(pi_project[0])
        project = clean_path(pi_project[1])

        # StudyDate
        if 'StudyDate' in dataset and len(dataset.StudyDate) > 0:
            StudyDate = dataset.StudyDate
        else:
            StudyDate = args.StudyDate

        study_date = clean_path(StudyDate)

        # PatientName
        if 'PatientName' in dataset and len(dataset.PatientName) > 0:
            PatientName = dataset.PatientName
        else:
            PatientName = args.PatientName

        patient = clean_path(PatientName.partition('^')[0])

        # studyID_and_hash_studyInstanceUID

        if 'StudyID' in dataset and len(dataset.StudyID) > 0:
            StudyID = dataset.StudyID
        else:
            StudyID = 'NA'

        if 'StudyInstanceUID' in dataset and len(dataset.StudyInstanceUID) > 0:
            StudyInstanceUID = dataset.StudyInstanceUID
        else:
            StudyInstanceUID = pydicom.uid.generate_uid()

        studyID_and_hash_studyInstanceUID = clean_path('.'.join([StudyID,
                                                                 hashcode(StudyInstanceUID)]))

        # SeriesNumber
        series_number = clean_path(
            '{series:04d}'.format(series=dataset.SeriesNumber))

        path = os.path.join(pi, project, study_date, patient,
                            studyID_and_hash_studyInstanceUID, series_number)
        sorted_filename = '{patient}.{modality}.{study}.{series:04d}.{image:04d}.{date}.{unique}.dcm'.format(
            patient=patient.upper(),
            modality=dataset.Modality,
            study=StudyDescription.upper(),
            series=dataset.SeriesNumber,
            image=dataset.InstanceNumber,
            date=StudyDate,
            unique=hashcode(dataset.SOPInstanceUID),
        )

        sorted_filename = clean_path(sorted_filename)

    except Exception as e:
        logger.exception('something wrong with {}'.format(filename))
        logger.exception(e)
        return None

    sorted_full_filename = os.path.join(path, sorted_filename)

    return sorted_full_filename


def sort_rule_clinical(filename):
    '''
    Clinical sort rule:

    patient_name
        |-study_date
            |-modality
                |-series_number
                  |-{patient}.{modality}.{series:04d}.{image:04d}.{study_date}.{unique}.dcm
                  ... 
                |-series_number
                ...
    intput:
        filename: dicom filename
    output:
        a dictionary:
            key: filename
            value: patient_name/study_date/modality/sereis_number/{patient}.{modality}.{series:04d}.{image:04d}.{study_date}.{unique}.dcm

    '''
    def write_error_file(my_file, errorInfoTemp):
        if os.path.exists(my_file):
            with open(my_file, 'r') as readFile:
                reader = csv.reader(readFile, delimiter='\t')
                lines = list(reader)
                if errorInfoTemp.split('\t') not in lines:
                    with open(my_file, 'a') as writeFile:
                        writeFile.write(errorInfoTemp)
                        writeFile.write("\n")
        else:
            with open(my_file, 'w') as writeFile:
                writeFile.write(
                    "\t".join(['subject', 'date', 'series', 'issue']))
                writeFile.write("\n")
                writeFile.write(errorInfoTemp)
                writeFile.write("\n")

    def clean_path(path):
        return re.sub(r'[^a-zA-Z0-9.-]', '_', '{0}'.format(path))

    def hashcode(value):
        code = 0
        for character in value:
            code = (code * 31 + ord(character)) & 0xffffffff
        return '{0:08X}'.format(code)

    # This will ignore any dicomdir present in the folder
    if all(['DICOMDIR' not in filename, not filename.endswith('OR_dates.tsv')]):
        logger = logging.getLogger(__name__)

        try:
            dataset = pydicom.read_file(
                filename, stop_before_pixels=True, force=True)
            study_date = dataset.StudyDate[0:4] + '_' + \
                dataset.StudyDate[4:6] + '_' + dataset.StudyDate[6:8]

            # This will skip any order sheets
            if dataset.Modality in {'SR', 'PR'}:
                errorInfoTemp = "\t".join(['P' + [s for s in filename.split('\\') if 'sub' in s][0].split('-')[1], study_date,
                                           clean_path('{series:04d}'.format(series=dataset.SeriesNumber)), dataset.Modality])
                my_file = os.path.join("\\".join(filename.split('\\')[:[i for i, s in enumerate(
                    filename.split('\\')) if 'sub' in s][0]]), 'errorInfo.tsv')
                write_error_file(my_file, errorInfoTemp)
                return None

            # This will skip any order sheets and localizers
            elif [x for x in dataset.ImageType if x in {'SECONDARY', 'LOCALIZER'}]:
                return None
            else:
                if 'SIEMENS' in dataset.Manufacturer:
                    errorInfoTemp = "\t".join(['P' + [s for s in filename.split('\\') if 'sub' in s][0].split('-')[1], study_date,
                                               clean_path('{series:04d}'.format(series=dataset.SeriesNumber)), 'SIEMENS'])
                    my_file = os.path.join("\\".join(filename.split('\\')[:[i for i, s in enumerate(
                        filename.split('\\')) if 'sub' in s][0]]), 'errorInfo.tsv')
                    write_error_file(my_file, errorInfoTemp)
                    return None
                else:
                    try:
                        csaReader = ds.wrapper_from_data(dataset)
                        modality = dataset.Modality

                        # --- INTRAOP X-RAY determination
                        if any(substring in modality for substring in {'Intraoperative', 'Skull', 'OT', 'XA', 'RF'}):
                            if 'CR' not in dataset.Modality:
                                or_date = dataset.StudyDate[0:4] + '_' + \
                                    dataset.StudyDate[4:6] + \
                                    '_' + dataset.StudyDate[6:8]
                                orDateTemp = "\t".join(
                                    ['P' + [s for s in filename.split('\\') if 'sub' in s][0].split('-')[1], or_date])
                                my_file = os.path.join("\\".join(filename.split('\\')[:[i for i, s in enumerate(
                                    filename.split('\\')) if 'sub' in s][0]+1]), 'OR_dates.tsv')

                                if os.path.exists(my_file):
                                    with open(my_file, 'r') as readFile:
                                        reader = csv.reader(
                                            readFile, delimiter='\t')
                                        lines = list(reader)
                                        if orDateTemp.split('\t') not in lines:
                                            with open(my_file, 'a') as writeFile:
                                                writeFile.write(orDateTemp)
                                                writeFile.write("\n")
                                else:
                                    with open(my_file, 'w') as writeFile:
                                        writeFile.write(
                                            "\t".join(['subject', 'or_date']))
                                        writeFile.write("\n")
                                        writeFile.write(orDateTemp)
                                        writeFile.write("\n")
                                return None

                            elif all(['CR' in dataset.Modality, 'Skull Routine Portable' in dataset.StudyDescription]):
                                errorInfoTemp = "\t".join(['P' + [s for s in filename.split('\\') if 'sub' in s][0].split('-')[1], study_date,
                                                           clean_path('{series:04d}'.format(series=dataset.SeriesNumber)), dataset.StudyDescription])
                                my_file = os.path.join("\\".join(filename.split('\\')[:[i for i, s in enumerate(
                                    filename.split('\\')) if 'sub' in s][0]]), 'errorInfo.tsv')
                                write_error_file(my_file, errorInfoTemp)
                                return None
                            else:
                                patient = 'P' + \
                                    [s for s in filename.split('\\') if 'sub' in s][0].split(
                                        '-')[1] + '_' + study_date
                                series_number = clean_path(
                                    '{series:04d}'.format(series=dataset.SeriesNumber))
                                studyID_and_hash_studyInstanceUID = clean_path('.'.join([dataset.StudyID or 'NA',
                                                                                         hashcode(dataset.StudyInstanceUID)]))

                                path = os.path.join(
                                    patient, dataset.StudyDate, studyID_and_hash_studyInstanceUID, modality, series_number)
                                sorted_filename = '{patient}.{modality}.{series:04d}.{image:04d}.{study_date}.{unique}.dcm'.format(
                                    patient=patient.upper(),
                                    modality=modality,
                                    series=dataset.SeriesNumber,
                                    image=dataset.InstanceNumber,
                                    study_date=dataset.StudyDate,
                                    unique=hashcode(dataset.SOPInstanceUID),
                                )
                        else:
                            if dataset.SeriesDescription.lower() not in {'loc', 'dose report'}:
                                patient = 'P' + \
                                    [s for s in filename.split('\\') if 'sub' in s][0].split(
                                        '-')[1] + '_' + study_date
                                series_number = clean_path(
                                    '{series:04d}'.format(series=dataset.SeriesNumber))
                                studyID_and_hash_studyInstanceUID = clean_path('.'.join([dataset.StudyID or 'NA',
                                                                                         hashcode(dataset.StudyInstanceUID)]))

                                path = os.path.join(
                                    patient, dataset.StudyDate, studyID_and_hash_studyInstanceUID, modality, series_number)
                                sorted_filename = '{patient}.{modality}.{series:04d}.{image:04d}.{study_date}.{unique}.dcm'.format(
                                    patient=patient.upper(),
                                    modality=modality,
                                    series=dataset.SeriesNumber,
                                    image=dataset.InstanceNumber,
                                    study_date=dataset.StudyDate,
                                    unique=hashcode(dataset.SOPInstanceUID),
                                )
                    except Exception as e:
                        errorInfoTemp = "\t".join(['P' + [s for s in filename.split('\\') if 'sub' in s][0].split('-')[1], study_date,
                                                   clean_path('{series:04d}'.format(series=dataset.SeriesNumber)), 'csaReader'])
                        my_file = os.path.join("\\".join(filename.split('\\')[:[i for i, s in enumerate(
                            filename.split('\\')) if 'sub' in s][0]]), 'errorInfo.tsv')
                        write_error_file(my_file, errorInfoTemp)
                        return None

        except Exception as e:
            logger.exception('something wrong with {}'.format(filename))
            logger.exception(e)
            return None

        if 'path' in locals():
            sorted_full_filename = os.path.join(path, sorted_filename)
            return sorted_full_filename
        else:
            return None



#!/usr/bin/env python
'''
sort or tar CFMM' data with DicomSorter


Update: 2019-11-21

Add DICOM tag option arguments to handel DICOM files missing(or empty): 
StudyDesciption, StudyDate, PatientName, StudyID, StudyInstanceUID,
SeriesNumber,InstanceNumber, and SOPInstanceUID. These tags are 
necessary to CFMM sort rule


Author: YingLi Lu
Email:  yinglilu@gmail.com
Date:   2018-05-22

Note:
    Tested on windows 10/ubuntu 16.04, python 2.7.14

'''
import sys
import os
import logging
import argparse

import sort_rules
import DicomSorter

import clinical_helpers as ch

logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s - %(levelname)s -%(message)s')


def main(dicom_dir, output_dir, args):
    '''
    use DicomSorter sort or tar CFMM's dicom data

    input:
        dicom_dir: folder contains dicom files(and/or compressed files:.zip/.tgz/.tar.gz/.tar.bz2)
        output_dir: output sorted or tar files to this folder
    '''

    logger = logging.getLogger(__name__)

    if not os.path.exists(dicom_dir):
        logger.error("{} not exist!".format(dicom_dir))
        return False

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    ######
    # CFMM sort rule
    ######
    try:
        if not args.clinical_scans:

            with DicomSorter.DicomSorter(dicom_dir, sort_rules.sort_rule_CFMM, output_dir, args) as d:
                # #######
                # # sort
                # #######
                # sorted_dirs = d.sort()
                # # logging
                # for item in sorted_dirs:
                #     logger.info("sorted directory created: {}".format(item))

                #######
                # tar
                #######
                # pi/project/study_date/patient/studyID_and_hash_studyInstanceUID
                tar_full_filenames = d.tar(5)
                # logging
                for item in tar_full_filenames:
                    logger.info("tar file created: {}".format(item))

            # ######
            # # demo sort rule
            # ######
            # with DicomSorter.DicomSorter(dicom_dir, sort_rules.sort_rule_demo, output_dir) as d:
            #     # sort
            #     sorted_dirs = d.sort()
            #     #logging
            #     for item in sorted_dirs:
            #         logger.info("sorted directory created: {}".format(item))

            #     # tar
            #     # patient_name/study_date/series_number/new_filename.dcm
            #     tar_full_filenames = d.tar(2)
            #     # logging
            #     for item in tar_full_filenames:
            #         logger.info("tar file created: {}".format(item))

        else:
            ######
            # Clinical sort rule
            ######
            logger.info("These are clinical scans.")

            with DicomSorter.DicomSorter(dicom_dir, sort_rules.sort_rule_clinical, output_dir, args) as d:
                # tar
                # study_date/patient/modality/series_number/new_filename.dcm
                tar_full_filenames = d.tar(4)

                # logging
                for item in tar_full_filenames:
                    logger.info("tar file created: {}".format(item))

            ch.tarSession(output_dir, dicom_dir, modality_sep=True)
            ch.combine_or_dates(dicom_dir, output_dir)
            ch.combine_error_info_tsv(dicom_dir, output_dir)

    except Exception as e:
        logger.exception(e)


def run():

    # arg parser
    parser = argparse.ArgumentParser()

    parser.add_argument("dicom_dir")
    parser.add_argument('output_dir')
    parser.add_argument("--clinical_scans", action="store_true")
    parser.add_argument("--StudyDescription",
                        nargs='?', default='PI^Project')
    parser.add_argument("--StudyDate",
                        nargs='?', default='19000101')
    parser.add_argument("--PatientName",
                        nargs='?', default='Anonymous')
    args = parser.parse_args()

    dicom_dir = args.dicom_dir
    output_dir = args.output_dir

    # main
    main(dicom_dir, output_dir, args)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
"""
Created on Mon May 13 14:03:24 2019

@author: User
"""
import os
import numpy as np
import pandas as pd
from tempfile import mkdtemp
import  tarfile
import shutil
import csv

def tarSession(tar_dir, source_dir, modality_sep):
    
    tar_files = [f for f in os.listdir(tar_dir) if f.endswith('.tar')]
    subjects = np.unique([x.split('_')[0] for x in tar_files])
    isub = 0
    for isub in range(len(subjects)):
        tar_files_sub = [f for f in os.listdir(tar_dir) if f.startswith(subjects[isub])]
        tar_files_sub_check = tar_files_sub[0].split('_')
        if len(tar_files_sub_check) > 7:
            print('Incorrect .tar filname for subject ' + subjects[isub])
            continue
        else:
            tar_files_dates = [x.split('_')[4] for x in tar_files_sub]
            date_sort_idx = np.argsort(tar_files_dates)
            
            tar_files_sub = [tar_files_sub[i] for i in date_sort_idx]
            tar_files_dates = [tar_files_dates[i] for i in date_sort_idx]
            
            sessionDates = {'date':[],'session':[]}
            session = 1
            for idate in range(len(tar_files_dates)):
                if modality_sep:
                    sessionDates['date'].append(tar_files_dates[idate])
                    sessionDates['session'].append(str(session).zfill(3))
                    session += 1
                else:
                    sessionDates['date'].append(tar_files_dates[idate])
                    sessionDates['session'].append(str(session).zfill(3))
                    session += 1
                    
            sessionDates = pd.DataFrame.from_dict(sessionDates)
            
            for i in range(len(tar_files_sub)):
                
                session = sessionDates.loc[i,'session']
                tmpdir = mkdtemp(prefix='DCM')
                
                tf = tarfile.open(os.path.join(tar_dir,tar_files_sub[i]))
                tmembers = tf.getmembers()
                for tm in tmembers:
                    tm.mode = 0o700
                tf_content = [m.name for m in tmembers if m.isfile()]
                fullTarPath = [os.path.join(tmpdir, f) for f in tf_content]
                tf.extractall(path=tmpdir, members=tmembers)
                tf.close()
                
                newName = tar_files_sub[i].split('_')
                newName = os.path.join(tar_dir, '_'.join([newName[0], session] + newName[1:]))
                
                with tarfile.open(newName, "w") as tar:
                    for filename in fullTarPath:
                        original_full_filename = filename
                        newName = filename.split('\\')[-1].split('/')
                        firstPart = '_'.join([newName[0].split('_')[0], session, newName[0].split('_')[1], newName[0].split('_')[2], newName[0].split('_')[3]])
                        lastPart = '_'.join([newName[-1].split('_')[0], session, newName[-1][5:]])
                        newName[0] = firstPart
                        newName[-1] = lastPart
                        arcname = '/'.join(newName)
                        tar.add(original_full_filename, arcname=arcname)
        
                shutil.rmtree(tmpdir)
                os.remove(os.path.join(tar_dir, tar_files_sub[i]))    
                
                print('Finished subject ' + subjects[isub] + ' session ' + session)

def combine_or_dates(source_dir, tar_dir):
    subs = [f for f in os.listdir(source_dir) if f.startswith('sub')]
    masterList = []
    for isub in range(len(subs)):
        path_to_sub = os.path.join(source_dir, subs[isub])
        my_file = os.path.join(path_to_sub, 'OR_dates.tsv')
        if os.path.exists(my_file):
            with open(my_file, 'r') as readFile:
                reader = csv.reader(readFile, delimiter='\t')
                next(reader, None)
                lines = list(reader)
            if len(lines)>1:
                sortedDates = [x[1] for x in lines]
                dateSort = np.argsort(sortedDates)
                for isort in range(len(dateSort)):
                    masterList.append(lines[dateSort[isort]])
            else:
                masterList.append(lines[0])
            
            os.remove(my_file) 
            
    if not masterList: 
        if os.path.exists(os.path.join(source_dir, 'or_dates.tsv')):
            with open(os.path.join(source_dir, 'or_dates.tsv'), 'r') as readFile:
                reader = csv.reader(readFile, delimiter='\t')
                next(reader, None)
                masterList = list(reader)
         
    final_file = os.path.join("\\".join(tar_dir.split('\\')[:-1]), 'or_dates.tsv')
    if os.path.exists(final_file):
        with open(final_file, 'r') as readFile:
            reader = csv.reader(readFile, delimiter='\t')
            next(reader, None)
            lines = list(reader)
        masterList = lines + [x for x in masterList if x not in lines]
        finalList = sorted(masterList, key=lambda x: (x[0], x[1]))
    else:
        finalList = masterList
                                        
    with open(final_file, 'w') as writeFile:
        writeFile.write("\t".join(['subject','or_date']))
        writeFile.write( "\n" )
        for i in range(len(finalList)):
            writeFile.write("\t".join(finalList[i]))
            writeFile.write( "\n" )
            
def combine_error_info_tsv(source_dir, tar_dir):
    if os.path.exists(os.path.join(source_dir, 'errorInfo.tsv')):
        with open(os.path.join(source_dir, 'errorInfo.tsv'), 'r') as readFile:
            reader = csv.reader(readFile, delimiter='\t')
            next(reader, None)
            masterList = list(reader)
        final_file = os.path.join("\\".join(tar_dir.split('\\')[:-1]), 'errorInfo.tsv')
        if os.path.exists(final_file):
            with open(final_file, 'r') as readFile:
                reader = csv.reader(readFile, delimiter='\t')
                next(reader, None)
                lines = list(reader)
            masterList = lines + [x for x in masterList if x not in lines]
            finalList = sorted(masterList, key=lambda x: (x[0], x[1], x[2]))
        else:
            finalList = masterList
                                            
        with open(final_file, 'w') as writeFile:
            writeFile.write("\t".join(['subject','date', 'series']))
            writeFile.write( "\n" )
            for i in range(len(finalList)):
                writeFile.write("\t".join(finalList[i]))
                writeFile.write( "\n" )
        
        os.remove(os.path.join(source_dir, 'errorInfo.tsv')) 


#!/usr/bin/env python
'''
Define a DicomSorter class, which can decompress, sort dicom files, and tar the sorted files, to a destination directory.

Author: YingLi Lu
Email:  yinglilu@gmail.com
Date:   2018-05-22

Note:
    Tested on windows 10/ubuntu 16.04, python 2.7.14
'''

import os
import tarfile
import zipfile
import shutil
import tempfile
import uuid
import logging
import subprocess
from collections import defaultdict

import pydicom
import extractCMRRPhysio


class DicomSorter():
    '''
    Extract compressed files(if any), sort dicom files, or tar the sorted, to a destination directory.

    Given a dicom directory:
        1. Find compressed files(if any) recursively, extract them temporally. Support formats:.  zip, .tgz, .tar.gz, tar.bz2
        2. Find dicom files recursively, sort(organizing and renaming) them, according to a given 'sort_rule_function',
        3. Sort or Tar the sorted dicom files to a destination directory.

    attributes:
        dicom_dir:
            dicom files directory, can have compressed files(.zip/.tgz/.tar.gz/.tar.bz2) under it.
        sort_rule_function:
            a function define how to sort(organizing and renaming) dicom files
        output_dir:
            save sorted dicom files or tar files
        extract_to_dir:
            extract compressed files to this directory temporally, default is platform's temp dir.
        dicomunwrap_path:
            path to dicomunwrap

    methods:
        tar()
        sort()

    Note:
        When extract larger compressed files on a platform, like sharcnet, with limit storage capacity temp folder,
        you might want to change the default 'extract_to_dir'

    '''

    def __init__(self, dicom_dir, sort_rule_function, output_dir, args,
                 extract_to_dir='', dicomunwrap_path='dicomunwrap', simens_cmrr_mb_unwrap_path='extractCMRRPhysio'):
        '''
        init DicomSorter
        '''
        self.logger = logging.getLogger(__name__)
        self._compressed_exts = ('.tar', '.tgz', '.tar.gz', '.tar.bz2', '.zip')
        self.dicom_dir = dicom_dir
        self.sort_rule_function = sort_rule_function
        self.output_dir = output_dir
        self.args = args

        # extract_to_dir, default is platform's tmp dir
        if not extract_to_dir:
            temp_dir = tempfile.gettempdir()
            self.extract_to_dir = temp_dir
        else:
            self.extract_to_dir = extract_to_dir

        # _extract_to_dir_uniq = extract_to_dir/uniq-string
        # `rmtree _extract_to_dir_uniq` after the sorting
        self._extract_to_dir_uniq = os.path.join(
            self.extract_to_dir, "DicomSorter_extract" + self._generate_uniq_string())
        if not os.path.exists(self._extract_to_dir_uniq):
            os.makedirs(self._extract_to_dir_uniq)

        # _unwrap_to_dir_uniq: wrap non-imaging data to this directory
        self._unwrap_to_dir_uniq = os.path.join(
            self.extract_to_dir, "DicomSorter_unwrap" + self._generate_uniq_string())
        if not os.path.exists(self._unwrap_to_dir_uniq):
            os.makedirs(self._unwrap_to_dir_uniq)

        self.dicomunwrap_path = dicomunwrap_path

        self.simens_cmrr_mb_unwrap_path = simens_cmrr_mb_unwrap_path

    def _generate_uniq_string(self):
        '''
        generate unique string
        '''
        return str(uuid.uuid4())

    def _check_non_imaging_and_unwrap(self, filename):
        '''
        check if the dicom file is non-imaging data(MRS, Physio,...), and unwrap it.

        input:
            filename: full path of dicom file

        output:
            None: if filename is a imaging dicom file
            a folder: if filename is a non-imaging dicom file

        '''
        try:
            dataset = pydicom.read_file(filename, stop_before_pixels=True)

            # dicomraw wrapped (wrapped by Igor's script)
            is_dicomraw_wrapped = \
                (0x0177, 0x0010) in dataset and \
                dataset[(0x0177, 0x0010)].value.startswith('Robarts^CFMM')

            # siemens CMRR MB sequence
            is_siemens_CMRR_MB_sequance = \
                "ImageType" in dataset and \
                dataset.ImageType == ['ORIGINAL', 'PRIMARY', 'RAWDATA', 'PHYSIO'] and \
                (0x7fe1, 0x0010) in dataset and \
                str(dataset[(0x7fe1, 0x0010)].value).strip(
                ) == 'SIEMENS CSA NON-IMAGE'

            output_directory = os.path.join(
                self._unwrap_to_dir_uniq, os.path.basename(filename))

            if not os.path.exists(output_directory):
                os.makedirs(output_directory)

            if is_dicomraw_wrapped:
                # unwrap command:
                # ./bin/dicomunwrap --input_file=/path/to/file.dcm --output_directory=/out/dir --decompress
                cmd = '{} '.format(self.dicomunwrap_path) +\
                    '--input_file={} '.format(filename) +\
                    '--output_directory={} '.format(output_directory) +\
                    '--decompress'

                subprocess.check_call(cmd, shell=True)
                return output_directory

            elif is_siemens_CMRR_MB_sequance:
                # unwrap command:
                # python extract_cmrr_physio.py  /path/to/file.dcm /out/dir
                cmd = '{} '.format(self.simens_cmrr_mb_unwrap_path) +\
                    '{} '.format(filename) +\
                    '{} '.format(output_directory)

                subprocess.check_call(cmd, shell=True)

                return output_directory

            else:
                return None

        except Exception as e:
            self.logger.exception(e)
            return None

    def sort(self):
        '''
        copy(organizing and renaming) dicom files into hierarchical directories

        output:
            tar_full_filename_list: list of resulted tar filenames

        note:
            write hierarchical direcotires and files on disk

        '''
        ######
        # extract compressed files if any
        ######
        self._walk_and_extract(
            self.dicom_dir, self._compressed_exts, self._extract_to_dir_uniq)

        # add _extract_to_dir_uniq directory in the search directories
        dicom_dirs = [self.dicom_dir, self._extract_to_dir_uniq]

        ######
        # walk and apply sort rule
        ######
        # return value is a list of list:
        #   [ [original_full_filename1, path/to/new-filename1],# [original_full_filename1, path/to/new-filename1],... ]
        before_after_sort_rule_list = self._walk_and_apply_sort_rule(
            dicom_dirs, self.sort_rule_function)

        # for logging
        sorted_dirs = []

        # copy: organizing and renaming original dicom files
        for item in before_after_sort_rule_list:

            # example: c:\\users\\user\\appdata\\local\\temp\\DicomSorter_8a46b089-fe90-4ee7-90fe-3cd9fc443d09\\0003.tar.gz816c904c-8e3e-4cff-8624-9fe4efd66815\\0003\\00001.dcm'
            original_full_filename = item[0]
            # example: PI\\Project\\19700101\\1970_01_01_T2\\1.9AC66A0D\\0003\\1970_01_01_T2.MR.PI_project.0003.0194.19700101.D6C44EC8.dcm
            relative_path_new_filename = item[1]

            # only the first element, example: PI
            sorted_dir = os.path.join(
                self.output_dir, relative_path_new_filename.split(os.sep)[0])
            if sorted_dir not in sorted_dirs:
                sorted_dirs.append(sorted_dir)

            full_path_new_full_filename = os.path.join(
                self.output_dir, relative_path_new_filename)
            dest_dir = os.path.dirname(full_path_new_full_filename)
            if not os.path.exists(dest_dir):
                os.makedirs(dest_dir)

            # copy imaging dicom
            shutil.copy(original_full_filename, full_path_new_full_filename)

            # check non-image diom and unwrap
            unwraped_dir = self._check_non_imaging_and_unwrap(
                original_full_filename)

            # copy unwraped dir
            if unwraped_dir:
                # create dir
                dest_unwraped_dir = full_path_new_full_filename+"_unwraped"

                # copytree's dst must not already exist
                if os.path.exists(dest_unwraped_dir):
                    shutil.rmtree(dest_unwraped_dir)

                # copy dir
                shutil.copytree(unwraped_dir, dest_unwraped_dir)

        return sorted_dirs

    def tar(self, depth, tar_filename_sep='_'):
        '''
        extract, apply sort rule, unwrap non-imaging dicom files, and create tar files(imaging->*.tar,non-imaging->*.attached.tar)

        input:
            depth: tar filename is named according to 'depth'.
                example:
                    given tree: /home/user/output_dir/project/study_date/patient_name/study_id
                    given depth = 4,
                    tar filename is: project_study_date_patient_name_study_id.tar
            tar_filename_sep: seprator of the tar file name elements

        output:
            tar_full_filename_list: list of resulted tar filenames

        note:
            write tar files on disk

        '''
        ######
        # extract compressed files if any
        ######
        self._walk_and_extract(
            self.dicom_dir, self._compressed_exts, self._extract_to_dir_uniq)

        # add _extract_to_dir_uniq directory in the search directories
        dicom_dirs = [self.dicom_dir, self._extract_to_dir_uniq]

        ######
        # walk and apply sort rule
        ######
        # return value is a list of list:
        #   [[original_full_filename1, path/to/new-filename1],[original_full_filename2, path/to/new-filename2],...]
        before_after_sort_rule_list = self._walk_and_apply_sort_rule(
            dicom_dirs, self.sort_rule_function)

        if not before_after_sort_rule_list:
            self.logger.info('dicom files no found!')
            return None

        ######
        # tar
        ######
        # tar imaging

        # get a dict, key is tar_full_filename, key is a list of list:
        #   {tar_full_filename1:[[original_full_filename1,/path/to/new_filename1],[original_full_filename2,/path/to/new_filename2],...],
        #   tar_full_filename2:[[original_full_filename1,/path/to/new_filename2],[original_full_filename2,/path/to/new_filename2],...],...}
        tar_full_filename_dict = defaultdict(list)

        for item in before_after_sort_rule_list:
            # c:\\users\\user\\appdata\\local\\temp\\DicomSorter_8a46b089-fe90-4ee7-90fe-3cd9fc443d09\\0003.tar.gz816c904c-8e3e-4cff-8624-9fe4efd66815\\0003\\00001.dcm'
            original_full_filename = item[0]
            # 'PI\\Project\\19700101\\1970_01_01_T2\\1.9AC66A0D\\0003\\1970_01_01_T2.MR.PI_project.0003.0194.19700101.D6C44EC8.dcm'
            relative_path_new_filename = item[1]

            # get tar file name
            # dir_split: ['PI','Project','19700101','1970_01_01_T2','1.9AC66A0D','0003','1970_01_01_T2.MR.PI_project.0003.0194.19700101.D6C44EC8.dcm']
            dir_split = relative_path_new_filename.split(os.sep)

            tar_filename = tar_filename_sep.join(dir_split[:depth])+".tar"
            tar_full_filename = os.path.join(self.output_dir, tar_filename)
            tar_full_filename_dict[tar_full_filename].append(item)

        for tar_full_filename, items in tar_full_filename_dict.items():
            with tarfile.open(tar_full_filename, "w") as tar:
                for item in items:
                    original_full_filename = item[0]
                    relative_path_new_filename = item[1]

                    arcname = relative_path_new_filename
                    tar.add(original_full_filename, arcname=arcname)

        # tar non-imaging:
        attached_tar_full_filenames = []
        for item in before_after_sort_rule_list:
            original_full_filename = item[0]
            relative_path_new_filename = item[1]

            unwraped_dir = self._check_non_imaging_and_unwrap(
                original_full_filename)

            if unwraped_dir:
                dir_split = relative_path_new_filename.split(os.sep)
                attached_tar_filename = tar_filename_sep.join(
                    dir_split[:depth])+".attached.tar"
                attached_tar_full_filename = os.path.join(
                    self.output_dir, attached_tar_filename)

                tar_arcname = relative_path_new_filename + "_unwraped"

                if attached_tar_full_filename not in attached_tar_full_filenames:
                    with tarfile.open(attached_tar_full_filename, "w") as tar:
                        tar.add(unwraped_dir, arcname=tar_arcname)

                    attached_tar_full_filenames.append(
                        attached_tar_full_filename)

                else:
                    with tarfile.open(attached_tar_full_filename, "a") as tar:
                        tar.add(unwraped_dir, arcname=tar_arcname)

        # for logging
        # list: compatible with python 3
        tar_full_filenames = list(tar_full_filename_dict.keys())

        return tar_full_filenames + attached_tar_full_filenames

    def _walk_and_apply_sort_rule(self, dicom_dirs, sort_rule_function):
        '''
        find each dicom files, apply sort rule

        input:
            dicom_dirs
            sort_rule_function:

        output:
            before_after_sort_rule_list: a list of list.
                sub_list[0]: original dicom-file's full path filename
                sub_list[1]: sorted-dicom-file's relative path filename: e.g. /pi/study_date/new-filename.dcm
        '''
        before_after_sort_rule_list = []
        for dicom_dir in dicom_dirs:
            for root, directories, filenames in os.walk(dicom_dir):
                for filename in filenames:
                    full_filename = os.path.join(root, filename)
                    if not full_filename.endswith(self._compressed_exts):
                        try:
                            sorted_relative_path_filename = sort_rule_function(
                                full_filename, self.args)
                            # apply sort_rule_function on non-dicom or bad dicom return None
                            if sorted_relative_path_filename is not None:
                                before_after_sort_rule_list.append(
                                    [full_filename, sorted_relative_path_filename])
                        except Exception as e:
                            self.logger.exception(e)

        return before_after_sort_rule_list

    def _extract(self, filename, to_dir):
        '''
        extract compressed files
        '''
        if (filename.endswith(".tar")):
            c_file = tarfile.open(filename, "r:")
        elif (filename.endswith(".tar.gz")) or (filename.endswith(".tgz")):
            c_file = tarfile.open(filename, "r:gz")
        elif (filename.endswith(".tar.bz2")):
            c_file = tarfile.open(filename, "r:bz2")
        elif (filename.endswith(".zip")):
            c_file = zipfile.ZipFile(filename)

        c_file.extractall(to_dir)
        c_file.close()

    def _walk_and_extract(self, input_dir, compressed_exts, to_dir):
        # self.input_dir, self._extract_to_dir_uniq

        # walk and extract
        for root, directories, filenames in os.walk(input_dir):
            for filename in filenames:
                full_filename = os.path.join(root, filename)
                if full_filename.endswith(compressed_exts):
                    # print 'decompressing ', full_filename
                    uniq_string = self._generate_uniq_string()
                    # filename+uniq_string: avoid same file names overwrite
                    output_dir = os.path.join(
                        to_dir, filename + uniq_string)

                    if not os.path.exists(output_dir):
                        os.makedirs(output_dir)

                    try:
                        self._extract(full_filename, output_dir)
                    except Exception as e:
                        self.logger.exception(e)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        '''
        remove temp directories
        '''
        if os.path.exists(self._extract_to_dir_uniq):
            shutil.rmtree(self._extract_to_dir_uniq)

        if os.path.exists(self._unwrap_to_dir_uniq):
            shutil.rmtree(self._unwrap_to_dir_uniq)
